<?php

namespace Oro\Bundle\FilterBundle\Filter;

use Symfony\Component\Form\FormFactoryInterface;

use Oro\Bundle\FilterBundle\Form\Type\Filter\DictionaryFilterType;
use Oro\Bundle\EntityBundle\Entity\Manager\DictionaryApiEntityManager;
use Oro\Bundle\EntityExtendBundle\Tools\ExtendHelper;
use Oro\Bundle\FilterBundle\Form\Type\Filter\EnumFilterType;
use Oro\Bundle\FilterBundle\Datasource\FilterDatasourceAdapterInterface;

/**
 * Enum filter provides possibility to create filter form and build an expression by passed enum data
 */
class EnumFilter extends BaseMultiChoiceFilter
{
    const FILTER_TYPE_NAME = 'enum';

    /** @var DictionaryApiEntityManager */
    protected $dictionaryApiEntityManager;

    /**
     * Constructor
     *
     * @param FormFactoryInterface $factory
     * @param FilterUtility        $util
     */
    public function __construct(
        FormFactoryInterface $factory,
        FilterUtility $util,
        DictionaryApiEntityManager $dictionaryApiEntityManager
    ) {
        parent::__construct($factory, $util);
        $this->dictionaryApiEntityManager = $dictionaryApiEntityManager;
    }

    /**
     * {@inheritdoc}
     */
    public function init($name, array $params)
    {
        $params[FilterUtility::FRONTEND_TYPE_KEY] = 'dictionary';
        if (isset($params['class'])) {
            $params[FilterUtility::FORM_OPTIONS_KEY]['class'] = $params['class'];
            unset($params['class']);
        }
        if (isset($params['enum_code'])) {
            $params[FilterUtility::FORM_OPTIONS_KEY] = [
                'enum_code' => $params['enum_code'],
                'class' => ExtendHelper::buildEnumValueClassName($params['enum_code'])
            ];
            $params['class'] = ExtendHelper::buildEnumValueClassName($params['enum_code']);
            unset($params['enum_code']);
        }

        parent::init($name, $params);
    }

    /**
     * {@inheritDoc}
     */
    public function getMetadata()
    {
        $metadata = parent::getMetadata();
        if ($metadata['class']) {
            $this->dictionaryApiEntityManager->setClass(
                $this->dictionaryApiEntityManager->resolveEntityClass($metadata['class'], true)
            );
            // Convert integer values in array to strings
            $keys = array_map('strval', $this->getForm()->get('value')->getData() ?? []);
            $metadata['initialData'] = $this->dictionaryApiEntityManager->findValueByPrimaryKey($keys);
        }

        return $metadata;
    }

    /**
     * {@inheritDoc}
     */
    protected function buildExpr(FilterDatasourceAdapterInterface $ds, $comparisonType, $fieldName, $data)
    {
        $parameterName = $ds->generateParameterName($this->getName());
        if (!in_array($comparisonType, [FilterUtility::TYPE_EMPTY, FilterUtility::TYPE_NOT_EMPTY], true)) {
            $ds->setParameter($parameterName, $data['value']);
        }

        return $this->buildComparisonExpr(
            $ds,
            $comparisonType,
            $fieldName,
            $parameterName
        );
    }

    /**
     * {@inheritdoc}
     */
    protected function getFormType()
    {
        return EnumFilterType::NAME;
    }

    /**
     * {@inheritdoc}
     */
    protected function buildComparisonExpr(
        FilterDatasourceAdapterInterface $ds,
        $comparisonType,
        $fieldName,
        $parameterName
    ) {
        switch ($comparisonType) {
            case DictionaryFilterType::TYPE_NOT_IN:
                return $ds->expr()->orX(
                    $ds->expr()->isNull($fieldName),
                    $ds->expr()->notIn($fieldName, $parameterName, true)
                );
            case DictionaryFilterType::EQUAL:
                return $ds->expr()->eq($fieldName, $parameterName, true);
            case DictionaryFilterType::NOT_EQUAL:
                return $ds->expr()->orX(
                    $ds->expr()->isNull($fieldName),
                    $ds->expr()->neq($fieldName, $parameterName, true)
                );
            default:
                return $ds->expr()->in($fieldName, $parameterName, true);
        }
    }
}
